#include <oxstd.h>
#include <oxprob.h>
#include <oxfloat.h>
#include <oxdraw.h>
#include <BankNames.ox>

#import <maximize>

static decl
	s_ammData,							// array of data
	s_mParameters,						// matrix of initial parameters
	s_ammParameters,					// pointer to array of parameter matrices  over time
	s_ammTrueParameters,				// pointer to array of true parameter matrices  over time
	s_mPosteriorMixtureProbabilities,	// matrix iN x iP of posterior mixture probabilities \tau_ij
	s_mU,s_mLogUConstant,				// matrix iT x iN x iP structure of Student t-auxilliary variables, and
										//    constant for the log(U) expectation
	s_dV,s_dVi,							// degrees of freedom in mv t-densities, and its inverse
	s_iP,								// # of mixture components
	s_dLik,								// the likelihood value
	s_iEstimateStatic,					// if 1, estimate a static model using initialization of mu
										// 		and Sigma as in E step
	s_iEstimateInitialMeans,			// if 1, we estimate the initial means; otherwise, use first XX weighted observations
	s_iEstimateInitialCovariances,		// if 1, we estimate the initial covariance matrices; otherwise, use first XX weighted observations
	s_dRegularizeCovariance = 0.00,	    // this number times a unit matrix is added to the covariance matrices to prevent degenerate results
	s_dFOffset,							// shift of the objective function
	s_iGlobalCount=0, 					// global count to track number of function evaluations
	s_iPrint,							// global indicator of amount of information to print
 	s_iRanSeed = 0,				   		// random seed
	s_mGASX = <>,						// explanatory covariates, such as TS factors
	s_mB1   = <>,						// coefficients B_1 for GASX specification
	s_mB1select = <>,					// indicating the non-zero entries in s_mB1.
	s_vIdt,								// a vector saving the SNL bank identifier
	s_ammPITTaus,						// an array with point-in-time s_mPosteriorMixtureProbabilities (based on log-densities at t)
	s_ammFilteredTaus,					// an array with filtered s_mPosteriorMixtureProbabilities (based on log-densities 0:t)
	s_sBankNames,						// sting with bank names for PIT diagnostic plots
	s_sName;							// string that selects path for saving files

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		             INITIALIZATION AND AUXILIARY ROUTINES                //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

fSwitch(const aA, const aB) {
// switch the contents of aA[0] and aB[0]
//
// INPUTS:
// aA, aB: pointers to objects
	decl aid1;
	aid1 = aA[0]; aA[0] = aB[0]; aB[0] = aid1;
}


fMaxr(const mX)	 {
//	OxMetrics 6 on HPC (High Performance Computing) does not know maxr command!
	decl r, vMaxr=zeros(mX[][0]);
	for (r=0;r<rows(mX);r++) vMaxr[r] = max(mX[r][]);
	return vMaxr;
}

fMinr(const mX)	 {
//	OxMetrics 6 on HPC (High Performance Computing) does not know maxr command!
	decl r, vMinr=zeros(mX[][0]);
	for (r=0;r<rows(mX);r++) vMinr[r] = min(mX[r][]);
	return vMinr;
}


fUseCommandLineArguments(const carglist, ...) {
// set the arguments equal to the command line arguments
	decl args = va_arglist();

	if (sizeof(carglist) == 1) return 0;
	if (sizeof(carglist) != sizeof(args)+1) {
		println("**** ERROR: ARGLIST != CARGLIST ***");
		println("Command line arguments:",carglist);
		println("Function arguments:",args);
		exit(0);
	}
	for (decl i1=1;i1<sizeof(carglist);++i1) if (sscan(carglist[i1],args[i1-1]) != 1) println("Error at ",i1,": ",carglist[i1]," and ",args[i1-1]);
}





fInitializeTimeVaryingParameterStructure(const iT, const iP, const iK, ...) {
// Initiale a 3d structure filled with zeros (or an optional scalar).
//
// INPUTS:
// iT	: size of the array
// iP	: rows of the array elements
// iK	: columns of the array elements
// dScalar: [optional] if provided, a scalar, otherwise set to 0
//
// OUTPUTS:
// iT x iP x iK structure filled with dScalar
//
	decl tmpParameters = new array[iT];
	decl args = va_arglist(), dScalar = 0;
	if (sizeof(args) > 0) dScalar = args[0];
	for (decl i1=0;i1<iT;++i1) tmpParameters[i1] = zeros(iP, iK) + dScalar;
	return tmpParameters;
}



fVecr3dStructure(const ammX) {
// Put the 3d structure ammX into vecr form, containing a ROW vector
// filled with vecr(ammParameters[0])'~...~vecr(ammParameters[iT])'.
//
// INPUTS:
// ammX: iT x iP x iK array.
//
// OUTPUTS:
// vP: 	1 x (iT * iP * iK) vector, containing vecr(ammParameters[0])'~...~vecr(ammParameters[iT])'
	decl vP = <>;
	for (decl i1=0;i1<sizeof(ammX);++i1) vP ~= vecr(ammX[i1])';
	return vP;
}


	

fReshape3dStructure(const vX, const iT, const iP, const iK) {
// Put the (iT*iP*iK) vector vX into a iT x iP x iK 3d structure mmX,
// where vX = fVecr3dStructure(mmX)
//
// INPUTS:
// vX: 1 x (iT*iP*iK) vector
// iT, iP, iK: desired output dimensions; 
//
// OUTPUTS:
// mmX: iT x iP x iK array.
	if (sizerc(vX) != iT*iP*iK) { println("dimension errors: size is ",sizerc(vX)," versus (",iT,",",iP,",",iK,")"); exit(0); }
	decl mmX = fInitializeTimeVaryingParameterStructure(iT,iP,iK);
	for (decl i1=0;i1<iT*iP*iK;i1+=iP*iK) mmX[i1] = reshape(vX[i1:i1+iP*iK-1],iP,iK);
	return mmX;
}


	

logistic(const vX) {
//Logistic transformation R -> (0,1)
return (vX .>= 0) .? (ones(vX) ./ (ones(vX)+exp(-vX))) .: (exp(vX) ./ (ones(vX)+exp(vX)));	}



logit(const mX) {
//Undoes logistic transformation (0,1) -> R
return log(mX) - log(ones(mX) - mX); }



RestrictPar(const vP) {
//restricts parameters to certain intervals
// implemented for one single smoothing parameter
	decl vPar = vP;
	vPar[0] = 0 + 2.0*logistic(vP[0]);
	vPar[1] = 0 + 1.0*logistic(vP[1]);
	//vPar[2:]= 0 + 2.0*(logistic(vP[2:])-0.5);
	//vPar[2:]= vP[2:];
	
	return vPar;
}


	
 
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        DRAWING AND PLOTTING ROUTINES                             //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

DrawParametric(const iWindow, const ammParameters, ...)
// Draws parameter paths in P=2 dimensions
{
	decl i1,i2,i3,y,x;	  //i1: time, i2: cluster
	decl iP = rows(s_mParameters);         // number of mixture components
	decl iT = sizeof(s_ammData);			// number of time points
	//decl iN = rows(s_ammData[0][][]);	    // number of cross sectional units
	decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	
	SetDrawWindow("DrawParametric");
	for (i3=0;i3<iD-1;i3+=2)				   //different variables
	for (i2=0;i2<iP;++i2)
	{
		y = x = <>;
		for (i1=0;i1<iT;++i1)
		{
			x ~= ammParameters[i1][i2][i3];	//component means in two-plane over time
			y ~= ammParameters[i1][i2][i3+1];
		}
		DrawXMatrix(iWindow+i3/2,y,0,x,0,0,1+i2);
	}
	ShowDrawWindow();
}

DrawMCov(const iWindow, const ammParameters, ...)
//Draws time-varying mcov parameters over time for iP clusters
{
	decl i1,i2,mChol, mCov,r;	 			//i1: time, i2: cluster
	decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	decl mAid;  
	
	SetDrawWindow("Time-varying se parameters");
	for (i2=0;i2<rows(ammParameters[0]);++i2)		 //over components 
	{	mAid = <>;
		for (i1=0;i1<sizeof(ammParameters);++i1)	//over time
		{ 	mChol = ammParameters[i1][i2][iD:iD+iD*(iD+1)/2-1];
			mChol = lower(unvech(mChol));
			mCov  = mChol * mChol';
			mAid ~= sqrt(diagonal(mCov))';	 //collect tv par over time
		}
		for (r=0;r<iD;r++) DrawTMatrix(iWindow+i2, mAid[r][], "tv std dev about tv means", 2008, 1, 4, 0, 1+r);
	}	
	ShowDrawWindow();
}



DrawTVpar(const iWindow, const ammParameters)
//Draws time-varying parameters over time in P=2 dimensions
{
	decl i1,i2;
	decl x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,xx1,xx2,xx3,xx4,xx5,xx6,xx7,xx8,xx9,xx10,xx11,xx12,xx13;  //i1: time, i2: cluster
	xx1 = xx2 = xx3 = xx4 = xx5 = xx6 = xx7 = xx8 = xx9 = xx10 = xx11 = xx12 = xx13 = <>;
	
	for (i2=0;i2<rows(ammParameters[0]);++i2)
	{
		x1 = x2 = x3 = x4 = x5 = x6 = x7 = x8 = x9 = x10 = x11 = x12 = x13 = <>;
		for (i1=0;i1<sizeof(ammParameters);++i1)
		{
			x1 ~= ammParameters[i1][i2][0]';
			x2 ~= ammParameters[i1][i2][1]';
			x3 ~= ammParameters[i1][i2][2]';
			x4 ~= ammParameters[i1][i2][3]';
			x5 ~= ammParameters[i1][i2][4]';
			x6 ~= ammParameters[i1][i2][5]';
			x7 ~= ammParameters[i1][i2][6]';
			x8 ~= ammParameters[i1][i2][7]';
			x9 ~= ammParameters[i1][i2][8]';
			x10~= ammParameters[i1][i2][9]';
			x11~= ammParameters[i1][i2][10]';
			x12~= ammParameters[i1][i2][11]';
			x13~= ammParameters[i1][i2][12]';
		}
		xx1 |= x1;
		xx2 |= x2;
		xx3 |= x3;
		xx4 |= x4;
		xx5 |= x5;
		xx6 |= x6;
		xx7 |= x7;
		xx8 |= x8;
		xx9 |= x9;
	   xx10 |= x10;
	   xx11 |= x11;
	   xx12 |= x12;
	   xx13 |= x13;
	}	

	SetDrawWindow("component means of transformed data "+sprint(s_iRanSeed));
	DrawTMatrix(iWindow+0,  xx1,  "component means, ln(total assets)"     , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+1,  xx2,  "ln(TA/CET1)"                           , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+2,  xx3,  "$\Phi^{-1}$(loans/assets)"             , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+3,  xx4,  "ln(risk mix)"                          , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+4,  xx5,  "trading assets / TA"                   , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+5,  xx6,  "derivatives / TA"                      , 2008, 1, 4, 0, 1);	
	DrawTMatrix(iWindow+6,  xx7,  "share of net interest income"          , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+7,  xx8,  "share of net fees and commissions"     , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+8,  xx9,  "share of trading income"               , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+9,  xx10, "retail loans/ loans"                  , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+10, xx11, "$\Phi^{-1}$(domestic loans/ loans)"    , 2008, 1, 4, 0, 1);
	//DrawTMatrix(iWindow+11, xx12, "loans/deposits ratio"                    , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+11, xx13, "ownership"                             , 2008, 1, 4, 0, 1);											   
	ShowDrawWindow();

	SetDrawWindow("component medians in orig scale "+sprint(s_iRanSeed));
	DrawTMatrix(iWindow+0, exp(xx1),   "component medians, TA (bn)"     , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+1, exp(xx2),   "leverage TA/CET1"               , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+2, probn(xx3), "loans/assets"                   , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+3, exp(xx4),   "risk mix"                       , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+4, xx5, "trading assets / total assets"         , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+5, xx6, "derivatives / total assets"            , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+6, xx7, "share of net interest income"          , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+7, xx8, "share of net fees and commissions"     , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+8, xx9, "share of trading income"               , 2008, 1, 4, 0, 1);
    DrawTMatrix(iWindow+9, xx10, "retail loans/ loans"                  , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+10,probn(xx11),"domestic loans/ loans"          , 2008, 1, 4, 0, 1);
	DrawTMatrix(iWindow+11,xx12,"loans/deposits ratio"                  , 2008, 1, 4, 0, 1);
	//DrawTMatrix(iWindow+11,xx13,"ownership"                             , 2008, 1, 4, 0, 1);				
	ShowDrawWindow();

}


DrawDiagnostic(const iWindow, const ammDiagnostic, const sTitle)
//draws posterior probabilities over time for a certain firm
{
	decl i2, i1, mX;
	decl vTauMax = zeros( rows(ammDiagnostic[0]) ,1);

	SetDrawWindow(sTitle+" posterior probabilites for firms 0-35 "+sprint(s_iRanSeed));
	for (i2=0;i2<36;++i2) {
		mX = <>;
		for (i1=0;i1<sizeof(ammDiagnostic);++i1) mX ~= ammDiagnostic[i1][i2][0:]';
		vTauMax[i2] = max(meanr(mX));
		DrawTMatrix(iWindow+i2, mX[:2][], s_sBankNames[i2], 2008, 1, 4, 0, 1);
		DrawTMatrix(iWindow+i2, mX[3:][], {}, 2008, 1, 4, 1, 1);
	}	
	ShowDrawWindow();
	
	SetDrawWindow(sTitle+" posterior probabilites for firms 36-71 "+sprint(s_iRanSeed));
	for (i2=36;i2<72;++i2) {
		mX = <>;
		for (i1=0;i1<sizeof(ammDiagnostic);++i1) mX ~= ammDiagnostic[i1][i2][0:]';
		vTauMax[i2] = max(meanr(mX));
		DrawTMatrix(iWindow+i2-36, mX[:2][], s_sBankNames[i2], 2008, 1, 4, 0, 1);
		DrawTMatrix(iWindow+i2-36, mX[3:][], {}, 2008, 1, 4, 1, 1);
	}	
	ShowDrawWindow();

	SetDrawWindow(sTitle+" posterior probabilites for firms 72-107 "+sprint(s_iRanSeed));
	for (i2=72;i2<108;++i2) {
		mX = <>;
		for (i1=0;i1<sizeof(ammDiagnostic);++i1) mX ~= ammDiagnostic[i1][i2][0:]';
		vTauMax[i2] = max(meanr(mX));
		DrawTMatrix(iWindow+i2-72, mX[:2][], s_sBankNames[i2], 2008, 1, 4, 0, 1);
		DrawTMatrix(iWindow+i2-72, mX[3:][], {}, 2008, 1, 4, 1, 1);
	}	
	ShowDrawWindow();

	SetDrawWindow(sTitle+" posterior probabilites for firms 108-143 "+sprint(s_iRanSeed));
	for (i2=108;i2<144;++i2) {
		mX = <>;
		for (i1=0;i1<sizeof(ammDiagnostic);++i1) mX ~= ammDiagnostic[i1][i2][0:]';
		vTauMax[i2] = max(meanr(mX));
		DrawTMatrix(iWindow+i2-108, mX[:2][], s_sBankNames[i2], 2008, 1, 4, 0, 1);
		DrawTMatrix(iWindow+i2-108, mX[3:][], {}, 2008, 1, 4, 1, 1);
	}	
	ShowDrawWindow();

	SetDrawWindow(sTitle+" posterior probabilites for firms 144-179 "+sprint(s_iRanSeed));
	for (i2=144;i2<180;++i2) {
		mX = <>;
		for (i1=0;i1<sizeof(ammDiagnostic);++i1) mX ~= ammDiagnostic[i1][i2][0:]';
		vTauMax[i2] = max(meanr(mX));
		DrawTMatrix(iWindow+i2-144, mX[:2][], s_sBankNames[i2], 2008, 1, 4, 0, 1);
		DrawTMatrix(iWindow+i2-144, mX[3:][], {}, 2008, 1, 4, 1, 1);
	}	
	ShowDrawWindow();

	SetDrawWindow(sTitle+" posterior probabilites for firms 180-end "+sprint(s_iRanSeed));
	for (i2=180;i2<208;++i2) {
		mX = <>;
		for (i1=0;i1<sizeof(ammDiagnostic);++i1) mX ~= ammDiagnostic[i1][i2][0:]';
		vTauMax[i2] = max(meanr(mX));
		DrawTMatrix(iWindow+i2-180, mX[:2][], s_sBankNames[i2], 2008, 1, 4, 0, 1);
		DrawTMatrix(iWindow+i2-180, mX[3:][], {}, 2008, 1, 4, 1, 1);
	}	
	ShowDrawWindow();

	//plot Histogram of (24)
	SetDrawWindow(sTitle+" Histogram of vTauMax "+sprint(s_iRanSeed));
		DrawDensity(0, vTauMax', "vTauMax", 1, 1, 1, 0, 0, 20, 2);
	ShowDrawWindow();

	return vTauMax;

}

DataMovie(const iDummy, ...) {
// Play data movie in Window 0
// INPUTS:
// iItem1		:	[optional] vector of integers indicating the characteristics to be plotted; by
//					default plot the first (max) five characteristics (or less if there are less)
// mPosteriorProbabilities : [optional] iN x iP array, iP number of mixture components,
//					containing the posterior probabilities; maximum probability indicates
//					color coding
// NO OUTPUTS
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics


	decl args = va_arglist();
	decl iItem1, mExtremes, vColors, vColorScheme = range(1,50);
	decl i1, i2, i3, i4, aid1;


	// set arguments (optional arguments)
	if (sizeof(args) > 0) {
		iItem1 = matrix(args[0]);
	} else {
		iItem1 = range(0,min(columns(s_ammData[0]),5)-1);
	}
	if (sizeof(args) > 1) {
		if (rows(args[1]) == rows(s_ammData[0])) vColors = vColorScheme[maxcindex(args[1]')]';
		else {
			println("\n\n*** ERROR: rows posterior probabilities not equal to ",rows(s_ammData[0])," ***\n");
			vColors = zeros(s_ammData[0][][0]) + vColorScheme[0];
		}
	} else {
		vColors = zeros(s_ammData[0][][0]) + vColorScheme[0];
	}

	// loop over time to determine the dimensions of the plots
	mExtremes = <>;
	for (i1=0;i1<sizeof(iItem1);++i1) {
		for (i2=0;i2<sizeof(s_ammData);++i2) {
			aid1 = deleter(s_ammData[i2][][iItem1[i1]]);
			if (i2>0) {
				mExtremes[i1][] =
					min(mExtremes[i1][0], minc(aid1)) ~
					max(mExtremes[i1][1], maxc(aid1)) ;
			} else {
				mExtremes |= minc(aid1) ~ maxc(aid1);
			}
		}
		mExtremes[i1][] += 0.05 * mExtremes[i1][] * <-1;1> * <-1,1>;
	}


	// loop over time to make the movie
	for (i1=0;i1<sizeof(s_ammData);++i1) {
		for (i2=minc(vColors);i2<=maxc(vColors);++i2) {
			for (i3=0;i3<sizeof(iItem1);++i3) {
				for (i4=0;i4<sizeof(iItem1);++i4) {

					if (i3==i4) continue;
					decl iWindow = i3*sizeof(iItem1) + i4;
			
					DrawXMatrix(iWindow,
						selectifr(s_ammData[i1][][iItem1[i3]],vColors.==i2)',0,
						selectifr(s_ammData[i1][][iItem1[i4]],vColors.==i2)',0,
						1,i2);
					DrawAdjust(ADJ_SYMBOL,2,150);
					DrawAdjust(ADJ_AREA_X,iWindow,mExtremes[i4][0],mExtremes[i4][1],(mExtremes[i4][1]-mExtremes[i4][0])/20);
					DrawAdjust(ADJ_AREA_Y,iWindow,mExtremes[i3][0],mExtremes[i3][1],(mExtremes[i3][1]-mExtremes[i3][0])/20);
				}
			}
		}
		DrawAdjust(ADJ_AREAMATRIX,sizeof(iItem1),sizeof(iItem1));
		ShowDrawWindow();
	}

}


	

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        ESTIMATION AND DIAGNOSTIC ROUTINES                        //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

InfoCriteria(const dlik, const iK, const iPrint)
{
	//numer of obs in the overall panel:
	decl iT = sizeof(s_ammData);			// number of time points
	decl iN = rows(s_ammData[0][][]);	    // number of cross sectional units
	decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	decl iP = rows(s_ammParameters[0]); 	// number of components
	decl iNum = iT * iN;                    // old command: sumr(sumc(s_mY .== .NaN .? 0 .: 1)); 

	decl AICc =  2*(iK - dlik) + 2*iK*(iK+1)/(iNum - iK - 1);
	decl BIC  = -2*dlik + iK * log(iNum);
	decl mIC  = dlik | iK | AICc | BIC;

	if (iPrint > 1) {
		println("********* AICc and BIC ********");
		println ("\nIC stats ",
	         "%c", {"model"},
			 "%r", {"loglik", "#Par", "AICc", "BIC"},
			 "%cf", {"%12.2f"}, mIC);
	}

	// Calculate Bai and Ng (2002)-type criteria	as well.
	// They might just work in this setting!
	decl i1, i2, dSSEw=0;
	decl mZ = (s_mPosteriorMixtureProbabilities .== fMaxr(s_mPosteriorMixtureProbabilities)*ones(1,iP) .? 1 .: 0);
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) dSSEw += sumc( sumsqrr(s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]).* mZ[][i2] );
	
	//compute Bai and Ng (2002) criteria, in log form. IC2 does best in SW's simulations.
	decl dOF  = dSSEw / (iN*iT*iD);
	decl IC1  = log(dOF) + iP * ((iN+iT)/(iN*iT)) * log((iN*iT)/(iN+iT));
	decl IC2  = log(dOF) + iP * ((iN+iT)/(iN*iT)) * log( fMinr(iN~iT) );
	decl IC3  = log(dOF) + iP * ( log(fMinr(iN~iT))/fMinr(iN~iT) );

	return {AICc, BIC, IC1, IC2, IC3};
		 
}

 

InfoCriteria2(const iPrint)	 {
	//numer of obs in the overall panel:
	decl i1, i2, i3, i4;
	decl iT = sizeof(s_ammData);			// number of time points
	decl iN = rows(s_ammData[0][][]);	    // number of cross sectional units
	decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	decl iP = rows(s_ammParameters[0]); 	// number of components

	decl mZ = (s_mPosteriorMixtureProbabilities .== fMaxr(s_mPosteriorMixtureProbabilities)*ones(1,iP) .? 1 .: 0);

	//calculate AICk, page 92.  
	decl dSSE=0;
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) dSSE += sumc( sumsqrr(s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]).* mZ[][i2] );
	dSSE /= iT;
	decl dAICk = dSSE + 2*(iN)*iP;


	// compute inverse data covariance matrix for Maha distance
	decl mcov = zeros(iD,iD);
	for (i1=0; i1<iT; i1++) mcov += variance(s_ammData[i1]); 			 
	mcov /=	iT;
	decl mcovi = invertsym(mcov);	//put to unit(iD) for Euclidian distances

	
	//calculate Calinski-Harabsaz Index, as in de Amorim and Hennig (2015), eq (3).
	decl mDiff, vDiff, vSelect, dSSEW=0;
	for (i1=0;i1<iT;++i1)
	for (i2=0;i2<iP;++i2)
	{	mDiff = s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1];
		vDiff = sumr((mDiff * mcovi) .* mDiff);
		vSelect = vDiff .* mZ[][i2];
		dSSEW += sumc( vSelect );
	}
	dSSEW /= iT;

	decl dSSEb=0;
	for (i1=0;i1<iT;++i1)
	for (i2=0;i2<iP;++i2)
	{ 	mDiff = s_ammParameters[i1][i2][:iD-1] - meanc(s_ammData[i1][][]);	//distance of estimated means from common data mean, a vector
		vDiff = sumr((mDiff * mcovi) .* mDiff);
		dSSEb += sumc(mZ[][i2])	* vDiff;
	}
	dSSEb /= iT;
	decl dCHK = (dSSEb / dSSEW) * ((iN - iP)/(iP-1));

	
	//calculate average Silhouette index, as in de Amorim and Hennig (2015), eq (2).
	decl dShC=0, vA, vB, vS, vDis, iNum;
	vS=vA=vB=zeros(iN,1);
	vDis=zeros(1,iP);
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iN;++i2) for (i3=0;i3<iP;++i3) if (mZ[i2][i3] == 1)  //this means: firm i2 is located in cluster i3.
		{	//average Mahalanobis distance of firm i2 to other firms in the same cluster i3
			mDiff = s_ammData[i1][][] - s_ammData[i1][i2][];
			vDiff = sumr((mDiff * mcovi) .* mDiff);
			vSelect = vDiff .* mZ[][i3];
			iNum    = sumc(  mZ[][i3] ) - 1;  //number of firms in cluster i3 other than firm i2
			vA[i2]  += ( 1/iNum ) * sumc( sqrt( vSelect ) );
			//now, calculate the average Euclidian distance of firm i2 to firms in other clusters i4
			for (i4=0;i4<iP;++i4)
			{ 	if (i4 == i3) vDis[i4] = .Inf; 
			  	if (i4 != i3)
				{	mDiff = s_ammData[i1][][] - s_ammData[i1][i2][];
					vDiff = sumr((mDiff * mcovi) .* mDiff);
					vSelect = vDiff .* mZ[][i4];
					iNum = sumc(mZ[][i4]);   //number of firms in cluster i4
					vDis[i4] = (1/ iNum) * sumc( sqrt( vSelect ) );
				};
			}
			//b_i is the shortest average distance to the most neighboring cluster
			vB[i2] += fMinr(vDis);
		}
	vA /= iT; vB /= iT;
	vS = (vB-vA) ./ fMaxr(vA~vB);
	
	
	//calculate Davies-Bouldin Index based on Mahalanobis distance. 
	decl vSi = zeros(1,iP), mM=zeros(iP,iP), mR=zeros(iP,iP);
	//compute time sum of cluster-specific scatters
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2)
	{	mDiff = s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1];
		vDiff = sumr((mDiff * mcovi) .* mDiff);
		vSelect = vDiff .* mZ[][i2];
		vSi[i2] +=	(1/sumc(mZ[][i2])) * sumc( sqrt ( vSelect ));
	};
	//compute time sum of distances between clusters 
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) for (i3=0;i3<iP;++i3)
	{	mDiff = s_ammParameters[i1][i2][:iD-1]	- s_ammParameters[i1][i3][:iD-1];
		vDiff = sumr((mDiff * mcovi) .* mDiff);
		mM[i2][i3] +=	sqrt ( vDiff );
	};
	//compute sum of two scatters / distance ratio for each i and j. Taking the ratio cancels iT. 
	for (i1=0;i1<iP;++i1) for (i2=0;i2<iP;++i2)	mR[i1][i2] = (vSi[i1] + vSi[i2]) / mM[i1][i2];

	mR = setdiagonal(mR, -.Inf);
	decl dDBI = meanc( fMaxr(mR) ); 

	
	//I return dSSE as well, to compute Hardigan value later in excel, as in de Amorim and Hennig (2015), eq (4). 
	return {-dCHK, dAICk, -meanc(vS), dDBI, dSSE};
		 
}


InitializeTau(const iNrMixtureComponents) {
// Initialize cluster means and covariance matrices.
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mParameters:iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// s_mPosteriorMixtureProbabilities: iN x iP matrix containing the mixture probabilities 
//				 for firm iN and mixture component iP
	decl iT = sizeof(s_ammData);
	decl iN = rows(s_ammData[0]);
	decl iD = columns(s_ammData[0]);
	decl mcov = zeros(iD,iD);
	decl vmu  = zeros(1, iD);

	// initialize tau's	as well
//	if (s_iRanSeed == 1079)
//	{	s_mPosteriorMixtureProbabilities  = ranchi(iN,iNrMixtureComponents,2);
//		s_mPosteriorMixtureProbabilities ./= sumr(s_mPosteriorMixtureProbabilities);
//	} else {	
//	decl mBestSoFar = loadmat(s_sName+"_BestTaus.xls")[][1:];
//	decl mPertubation = ranchi(iN,iNrMixtureComponents,2);					//pertube randomly
//	mPertubation ./= sumr(mPertubation);   
//	s_mPosteriorMixtureProbabilities = 2* mPertubation + 1* mBestSoFar;
//	s_mPosteriorMixtureProbabilities ./= sumr(s_mPosteriorMixtureProbabilities);
//  	}

	s_mPosteriorMixtureProbabilities = loadmat(s_sName+"_BestTaus_LSS.xls")[][1:];
	  
//	// compute initial means and covariances
//	mcov += variance(s_ammData[0] ); 			 
//	vmu  += meanc(   s_ammData[0] );
//
//	// randomize the starting value
//	s_mParameters = <>;
//	//for (decl r=0;r<iNrMixtureComponents;r++) s_mParameters |= ( 1.64*sqrt(diagonal(mcov)) .* rann(1,iD) ) .* vmu ~ 2*ranu(1,iD*(iD+1)/2).*vech(choleski(mcov))';
//	for (decl r=0;r<iNrMixtureComponents;r++) s_mParameters |=  vmu ~ vech(choleski(mcov))';

	

} 
	
	  


fInitializeVp(const mParameters, const iD, const dInverseDegreesOfFreedom, const iEstmateInitialMeans, const iEstimateInitialCovariances) {
	decl vP = <>;
	
	// single mean smoothing parameter
	vP |= -0.098387749 + 0*rann(1,1);									  		

	
	// no covariance smoothing parameter
	vP |= 0.471014439 + 0*rann(1,1); // = <>;

	// starting means
	if (iEstmateInitialMeans == 1) vP |= vecr(mParameters[][:iD-1]);

	// starting covariances
	if (iEstimateInitialCovariances== 1) vP |= vecr(mParameters[][iD:]);

	// degrees of freedom
	if (dInverseDegreesOfFreedom > 0) vP |= -0.604583103;

	//parameters for B_1 GASX.
	if (rows(s_mGASX)>0)	vP |= 1*rann( sumr(sumc(s_mB1select)), 1);	//add (#explanatory variables * iD) GASX coefficients

	return vP;
}





fGetParameters(const vP, const iNrMixtureComponents, const iEstimateInitialMeans, const iEstimateInitialCovariances) {
// Get parameters for the filter from the parameter vector, accounting for several options
//
// INPUTS:
// vP:			parameter vector (dimensions may vary)
// ammData:     iT x iN x iD array, with iN the number of cross-sectional units,
//              and iT the number of time points, and iD the number of characteristics
// iNrMixtureComponents: integer, the number of mixture components
// mPosteriorMixtureProbabilities: (iN x iP) matrix with iN units and iP regimes
// dVi:			inverse degrees of freedom, and degrees of freedom. If dVi==0 or dVi=.NaN, then Gaussian.
//					If dVi<0, the parameter is estimated.
// iEstimateInitialMean: if == 1, vP contains elements of initial means;
//				otherwise use initial XX weigthed observations
// iEstimateInitialCovariance: if == 1, vP contains elements of initial covariance matrices;
//				otherwise use initial XX weigthed observations
//
// OUTPUTS:
// vEwma: iD+1 x 1 parameter with iD smoothing parameters for the means, and 1 smoothing parameter for variances
// mNewParameters: iP x 0.5*iD*(iD+3) matrix with starting means for each  of the iP regimes, and
//				starting vech(chol(covariances))
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mPosteriorMixtureProbabilities: iN x iP matrix containing the mixture probabilities 
//				 for firm iN and mixture component iP
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated.


	decl ix=0;
 	decl iD = columns(s_ammData[0]);
	decl iP = s_iP;

	decl vPar = RestrictPar(vP);
	
	// first element(s) vP: smoothing parameter(s) for the mean
	decl vEwma = (ones(iD-1,1)|0) * vPar[ix];   //(vP[ix:ix+iD-2]|0);					  //last variable `ownership' is static!
	ix++;                             			//iD-1;
	
	// one additional element for smoothing parameter(s) for the covariance matrix
	vEwma |= vPar[ix];
	ix++;   

	// next elements vP: the initial means
	decl XX = minc(1 | sizeof(s_ammData));
	decl mNewParameters = zeros(iP,0.5*iD*(iD+3)), mCovariance, mChol;
	if (iEstimateInitialMeans == 1) {
		mNewParameters[][:iD-1] = reshape(vP[ix:ix+iP*iD-1],iP,iD); ix += iP * iD;
	} else {
		// update starting means and starting variances or put them to parameter values
		// put them equal to the weighted means of the first XX observations
		// weights used are of the Gaussian, **not** the Student's t
		for (decl t=0;t<XX;t++)	mNewParameters[][:iD-1] += s_mPosteriorMixtureProbabilities's_ammData[t][][];
		mNewParameters[][:iD-1] ./= XX * sumc(s_mPosteriorMixtureProbabilities)';
	}

	// next elements vP: the initial covariances
	if (iEstimateInitialCovariances == 1) {
		mNewParameters[][iD:] = reshape(vP[ix:ix+iP*(iD*(iD+1))/2-1],iP,(iD*(iD+1))/2); ix += iP * (iD*(iD+1))/2;
	} else {
		// update starting means and starting variances or put them to parameter values
		// put them equal to the weighted means of the first XX observations
		// weights used are of the Gaussian, **not** the Student's t
		for (decl r=0;r<iP;r++) {
			mCovariance = 0;
			for (decl t=0;t<XX;t++) mCovariance += (s_ammData[t][][] - mNewParameters[r][:iD-1])'*(s_mPosteriorMixtureProbabilities[][r].*(s_ammData[t][][] - mNewParameters[r][:iD-1]));
			mCovariance /= XX * sumc(s_mPosteriorMixtureProbabilities[][r]);
			mChol = choleski(mCovariance);
			mNewParameters[r][iD:] = vech(mChol)';
		}
	}

	// next element vP: degrees of freedom
	if (ismissing(s_dVi) || (s_dVi == 0)) {
		s_dV = .NaN;
	} else if (s_dVi > 0) {
		s_dVi = 0.5 * logistic(vPar[ix]); ix+=1;
		s_dV  = 1.0 / s_dVi;
	} else {
		s_dV = 1.0 / fabs(s_dVi);
	}

	if (rows(s_mGASX)>0)
	{	s_mB1 = zeros(rows(s_mGASX),iD);
		for (decl r=0;r<rows(s_mB1);r++) for (decl c=0;c<columns(s_mB1);c++) if (s_mB1select[r][c] == 1) {s_mB1[r][c] = vPar[ix]; ix++; };
	}
	return {vEwma, mNewParameters};
}



fGasFilter(mParameters, const vEwma, const iEstep) {
// Run the Gaussian GAS filter for the mean and the covariance matrix.
// At the moment, the filter is a single parameter EWMA filter.
//
// INPUTS:
// mParameters : iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// vEwma       : (iD+1)x1 double, the GAS smoothing parameter for mean (0..iD-1) and variance (iD element)
// iEstep	   : if iEstep == 1, update s_mU while going. Otherwise, do not update s_mU.
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mParameters:iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// s_mPosteriorMixtureProbabilities: iN x iP matrix containing the mixture probabilities 
//				 for firm iN and mixture component iP
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated. These are the values from the previous iteration,
//					needed for example when running the filter itself in the E-step. For the computation of likelihoods and
//					the objective function, however, we need the degrees of freedom of the current iteration.
//					This is particularly relevant for the M-step.
// s_mU			: iT x iN x iP structure, on entry should be there for the M step, and is updated if iEstep==1
//
// OUTPUTS:
// vELik       :  1 x iP vector holding the conditionally expected log likelihoods for each mixture component
// mLik        :  iN x iP matrix holding the log densities for each mixture component


	decl iD,iN,iT,iP,i1,i2,i,ix;
	decl vMean, mChol, mCov, mCovi, mCholOld, vELik, mLik, mU, mLogU, Score, dAlphaCov, mAidCov, mAidChol, mLikPrevious;
	decl vMixtureProbabilities = meanc(s_mPosteriorMixtureProbabilities);

	// initialize
	iT = sizeof(s_ammData);			// number of time points
	iN = rows(s_ammData[0][][]);	// number of cross sectional units
	iD = columns(s_ammData[0][][]);	// number of characteristics in data
	iP = rows(mParameters);         // number of mixture components
	vELik = zeros(1,iP);
	mLik = zeros(iN,iP);
	mU = ones(iN,1);

	
	// run the filter per time step
	for (i1=0;i1<iT;++i1) {

		// store parameters
		s_ammParameters[i1] = mParameters;
		mLikPrevious        = mLik;
		
		// loop over the mixture components
		for (i2=0;i2<iP;++i2) {
		
			// get out the parameters
			ix = 0;
			vMean = mParameters[i2][ix:ix+iD-1];	//vMean refers to component i2 at time i1, of dimension iD x iD
			ix += iD;
			mChol = mParameters[i2][ix:ix+iD*(iD+1)/2-1];
			ix += iD*(iD+1)/2;
			mChol = lower(unvech(mChol));
			mCholOld = mChol;
			
			mCov = mChol * mChol' + s_dRegularizeCovariance * unit(iD);
			mCovi = invertsym(mCov);
			
			
			// compute the new estimate of the unconditional mixture probabilities (pi-s)
			// add the time-t log density contribution for component j (for each firm); the f_j part for (A1)
			if (ismissing(s_dVi) || (s_dVi == 0)) {
				// Gaussian case
				mLik[][i2] += - 0.5*iD*log(M_2PI)											
							 - sumr(log(diagonal(mChol)))
							 -  sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean));
			} else {
				// Student's t case
				// compute conditionally expected likelihood value 
				mLik[][i2] +=  loggamma( (s_dV + iD)/2 )
							 - loggamma( s_dV/2 )
							 - 0.5*iD*log(s_dV * M_PI)											
							 - sumr(log(diagonal(mChol)))
							 - 0.5 * (s_dV + iD) * log(1 + fabs(s_dVi) * sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean)));
			}


			// compute conditionally expected likelihood value 
			// Add Q1 part: sum(i, tau(ij) * log(pi(j))); add only for one
			if (i1==0) {
				// time point (as cluster assignments are time-invariant)
				vELik[i2] = log(vMixtureProbabilities[i2]) * sumc(s_mPosteriorMixtureProbabilities[][i2]);
			}
			// Add Q2 part for Gaussian or Student's t case
			if (ismissing(s_dVi) || (s_dVi == 0)) {
				// Gaussian case
				vELik[i2] += s_mPosteriorMixtureProbabilities[][i2]'
						(
							- 0.5 * iD * log(M_2PI)
							- sumr(log(diagonal(mChol)))
							- 0.5 * sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean))
						);
				mLogU = ones(iN,1);
			} else {
				// Equation (A4)-type for fixed (j,t), summed over i
				vELik[i2] +=	s_mPosteriorMixtureProbabilities[][i2]'
						(
							(loggamma((s_dV+iD)/2) - loggamma(s_dV/2) - 0.5*iD*log(s_dV * M_PI))
							- sumr(log(diagonal(mChol)))
							- 0.5 * (s_dV + iD) * log(1 + fabs(s_dVi) * sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean)))
						);
				mLogU = (1 + fabs(s_dVi) * iD) ./ (1 + fabs(s_dVi) * sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean)));
			}

			//New: For diagnostic statistics, compute the point-in-time posterior probabilities at each time i1
			decl mLikt = mLik - mLikPrevious;
			decl dConst = fMaxr(mLikt);
			decl vWeightedDens = vMixtureProbabilities .* exp(mLikt - dConst);
			s_ammPITTaus[i1] = vWeightedDens ./ sumr(vWeightedDens);
			

			//New: Also compute the filtered posterior probabilities at each time i1
			dConst = fMaxr(mLik);
			vWeightedDens = vMixtureProbabilities .* exp(mLik - dConst);
			s_ammFilteredTaus[i1] = vWeightedDens ./ sumr(vWeightedDens);
			
			
////			// OLD CODE
////			// compute the score step weights
////			mU = s_mPosteriorMixtureProbabilities[][i2];
////			mU .*= mLogU / sumc(mU);
////			// compute the mean's score step and update the mean
////			Score = mU'(s_ammData[i1][][] - vMean);
////			vMean = vMean + vEwma[0:iD-1]' .* Score;
////			// compute the covariance matrix's score step and update the covariance matrix
////			dAlphaCov = (vEwma[iD]);
////			Score = (s_ammData[i1][][] - vMean)' * (mU .* (s_ammData[i1][][] - vMean)) - mCov;

			// compute the score step weights
			decl mTaus = s_mPosteriorMixtureProbabilities[][i2];
			decl mTausTimesUs = mTaus .* mLogU;
			// compute the mean's score step and update the mean
			decl vReg = zeros(1,iD);
			if (rows(s_mGASX)>0) for (decl r=0; r<rows(s_mGASX); r++) vReg += s_mB1[r][] * s_mGASX[r][i1]; //iD coefficients * common variable at that time

			Score = (mTausTimesUs)'(s_ammData[i1][][] - vReg - vMean) / sumc(mTaus);
			vMean = vMean + vEwma[0:iD-1]' .* Score;
			
			
			// compute the covariance matrix's score step and update the covariance matrix
			dAlphaCov = (vEwma[iD]);
			Score = (s_ammData[i1][][] - vMean)' * (mTausTimesUs .* (s_ammData[i1][][] - vMean))/sumc(mTaus) - mCov;
			for (i=0; i<10;i++) 
			{	//print(i);
				mAidCov  = mCov + dAlphaCov * Score; //update all  * diagcat(unit(iD),<>) 
				mAidChol = choleski(mAidCov);
				if (rows(mAidChol) > 1) mChol = mAidChol;  		//returned a proper matrix; thus update. If not, mChol remains what is was. 				
			};
			
			// store parameters into matrix object		 
			ix = 0;
			mParameters[i2][ix:ix+iD-1] = vMean;
			ix += iD;
			//mChol = choleski(mCov);
			//if (mChol == 0) mChol = mCholOld;
			mParameters[i2][ix:ix+iD*(iD+1)/2-1] = vech(mChol)';
			ix += iD*(iD+1)/2;

		}
	}

	// plot the time varying parameter path
//	DrawParametric(0, s_ammParameters); ShowDrawWindow();

	return {vELik, mLik};
}


fPerformEStep(const mParameters, const vEwma) {
// Compute the posterior probabilities tau_ij, i.e., do the E step of the EM algorithm for the mixture.
//
// INPUTS:
// mParameters : iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of cov matrix)'
// vEwma       : (iD+1)x1 double, the GAS smoothing parameter for mean (0..iD-1) and variance (iD element)
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mPosteriorMixtureProbabilities : pointer to iN x iP matrix containing, on entry containing
//               the posterior mixture probabilities for firm iN and mixture component iP.
//               On exit: containing the updated estimates of these.
// s_dVi		: contains the inverse degrees of freedom parameter (0 for Gaussian). If negative, it is a
//					pre-set parameter, if positive, the parameter is estimated
//
// NO OUTPUTS
               
	decl vMixtureProbabilities, vWeightedDens, mLik, vELik;

	// compute new prior probabilities \pi from old \tau_ij's
	vMixtureProbabilities = meanc(s_mPosteriorMixtureProbabilities);   		 //eq 7 scribbles

	// run the filter at the latest parameters
	[vELik, mLik] = fGasFilter(mParameters, vEwma, 1);

	// compute the new posterior probabilities
	s_dLik = fMaxr(mLik);
	vWeightedDens = vMixtureProbabilities .* exp(mLik - s_dLik);
	s_dLik = double(sumc(log(sumr(vWeightedDens)) + s_dLik));
	decl mCandidateMixtureProbabilities = vWeightedDens ./ sumr(vWeightedDens);

	 //update tau's only if at least iD /iN or x% of firms in each cluster 
	decl iN = rows(s_ammData[0]);
	decl iD = columns(s_ammData[0]);
	if ( fMinr(meanc( mCandidateMixtureProbabilities )) >= (0.05)  ) {
		s_mPosteriorMixtureProbabilities = mCandidateMixtureProbabilities;
	} else {
		println("*** No improvement on posterior weights E-step ***");
	}

}




fMaximizeStep1(const vP, const adF, const avG, const amH)
{
// Perform the maximization step and return the conditionally expected
// log likelihood function.
// This routine has the input/output format suited for the MaxBFGS optimizer.
//
// Data: should be known in the array iT x iN x iD variable s_ammData
//
// THIS VERSION INPUT vP:
// estimate the single smoothing parameteter for the mean,
// the initial means, and
// all of the mixture component covariance matrices matrices
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mParameters:iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// s_mPosteriorMixtureProbabilities: iN x iP matrix containing the mixture probabilities 
//				 for firm iN and mixture component iP
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated.
// s_iEstimateInitialMeans: if 1, we estimate the initial means; otherwise, use first XX weighted observations
// s_iEstimateInitialCovariances: if 1, we estimate the initial covariance matrices; otherwise, use
//				 first XX weighted observations
//


	decl ix,iP,iN,iD,iT;
	decl vEwma;

	// get the dimensions
	iT = sizeof(s_ammData);
	iN = rows(s_ammData[0]);
	iD = columns(s_ammData[0]);
	iP = s_iP;
	if (fmod(iP,1) != 0) {
		println("*** ERROR in fMaximizeStep1(..): dimensions of vP[",rows(vP),"] incorrect ***");
		println("iT = ",iT,"\niN = ",iN,"\niD = ",iD,"\niP = ",iP);
		exit(0);
	}

	// get the parameters
	[vEwma, s_mParameters] = fGetParameters(vP, iP, s_iEstimateInitialMeans, s_iEstimateInitialCovariances);

	// run the filter
	decl ELik, Lik;
	[ELik, Lik] = fGasFilter(s_mParameters, vEwma, 0);

	
	adF[0] = double(sumr(ELik) - s_dFOffset);
	if (ismissing(adF[0])) return 0; else return 1;
}


ParStderr(const vP)
{
	println("Computing parameter standard error...");
    decl covar, invcov, result;

    result = Num2Derivative(fMaximizeStep1, vP, &covar);

	if (!result)
    {
		print("Num2Derivative() failed in numerical second derivatives\n");
        return zeros(vP);
    }
    invcov = invertgen(-covar, 30);
    decl se =  sqrt(diagonal(invcov))';

	decl ip = 0, m, v;
		println("Parameter estimates (coeff - res - s.e. - t-value)");
		println(vP ~ RestrictPar(vP) ~ se  ~ fabs(vP ./ se));
		//savemat(s_spec+"vPar_SE.in7",  vP ~ RestrictPar(vP) ~ se  ~ (vP ./ se)  );
	
	return {se, (vP ./ se) };
}


fRunEmAlgorithm(const iNrMixtures, const iNrStartingValues, const iPrint, const dInverseDegreesOfFreedom)
{
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mParameters:iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// s_mPosteriorMixtureProbabilities: iN x iP matrix containing the mixture probabilities 
//				 for firm iN and mixture component iP
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated.
// s_iEstimateInitialMeans: if 1, we estimate the initial means; otherwise, use first XX weighted observations
// s_iEstimateInitialCovariances: if 1, we estimate the initial covariance matrices; otherwise, use
//				 first XX weighted observations
	decl iMaxEMIterations = 50;
	decl iEmConverged = 0;
	decl iP = iNrMixtures;
	decl iD = columns(s_ammData[0]);
	decl iN = rows(s_ammData[0]);
	decl iT = sizeof(s_ammData);
	decl vP, dAICc, dBIC, dBNg1, dBNg2, dBNg3, dNCH, dAICk, dNSC, dSSEw, dDBI, dF=0, vEwma;

	// initialize
	s_iP = iP;
	// shift in objective function (numerical issue)
	s_dFOffset = 0;
	
	//Iterate E and M steps:
	decl i, ir, vSave, dFold=1, dconv=1, vTauMax, mResults = <>, mResults2 = <>;
	if (iPrint > 4) MaxControl(250, 1, 0); else MaxControl(250, 10, 1);
//	MaxControlEps(1e-4, 1e-3);
	

	// loop over number of staring values
	for (decl iS=0;iS<iNrStartingValues;iS++) {
	//for (decl iS=0;iS<1;iS++)	{  
		// initialize random taus, parameters
		println("\n *** Start iteration "+sprint(iS));

		s_iRanSeed = 0 + iS;	//4776;	 389;  1839; 12890
		ranseed(s_iRanSeed);
		s_dVi = dInverseDegreesOfFreedom;
		InitializeTau(iNrMixtures);
		vP = fInitializeVp(s_mParameters, iD, dInverseDegreesOfFreedom, s_iEstimateInitialMeans, s_iEstimateInitialCovariances);
		
		for (decl i0=0;i0>-1;--i0) {	  //two rounds: estimates first a static model, then the dynamic model. i0=1 two rounds, i0=0 one round.
	
			dconv = 1;
			s_iEstimateStatic = i0;
		
			for (i = 0; i < iMaxEMIterations; i++) {
			
				//E-Step, producing s_mPosteriorMixtureProbabilities 
				if (iPrint > 1) println("E-Step"+sprint(i));
				[vEwma, s_mParameters] = fGetParameters(vP, iP, s_iEstimateInitialMeans, s_iEstimateInitialCovariances);
				//println("s_mParameters[][:5]: ", s_mParameters[][:5]);
				fPerformEStep(s_mParameters, vEwma);
				s_mPosteriorMixtureProbabilities = loadmat(s_sName+"_BestTaus_LSS.xls")[][1:]; //prevent EStep

				//M-Step
				if (iPrint > 1) println("M-Step"+sprint(i));
				ir = MaxBFGS(fMaximizeStep1, &vP, &dF, 0, 1);	  

				println("vP': ", vP');
				
				// check convergence
		 		if (i>1) dconv = max(fabs((dF - dFold) ./ dFold));
				if (dconv < 10^-5) {iEmConverged = 1; break;}
				dFold = dF;

			}

			if (iPrint > 0) 
				println("\nEM ",(iEmConverged==1)?"":"not ","converged in ",i," iterations");
			if (iPrint > 1) {
				println("vP unrestricted, restricted: ", vP ~ RestrictPar(vP) );
				//println("s_mParameters: ", s_mParameters);
				println("pi's are: ", meanc(s_mPosteriorMixtureProbabilities) );
			}
			if (iPrint > 0) 
				println("\n======================================================\n\n");

			if (s_iPrint > -10)
			{	//DrawParametric(0, s_ammParameters);
				DrawTVpar(0, s_ammParameters);
				DrawMCov(0, s_ammParameters);
				DrawDiagnostic(0, s_ammPITTaus, "PIT");
				vTauMax = DrawDiagnostic(0, s_ammFilteredTaus, "Filtered");
				
				
			}	
		
		} //end estimation step
	
		//report info criteria
		decl iK = rows(vP)	                //alpha + possibly a df parameter
				  + sizerc(s_mParameters) 	//all initial means and covariance parameters
				  + columns(s_mPosteriorMixtureProbabilities); //pi mixture probabilities
					
		[dAICc, dBIC, dBNg1, dBNg2, dBNg3] = InfoCriteria(s_dLik, iK, iPrint);
		[dNCH, dAICk, dNSC, dDBI, dSSEw]   = InfoCriteria2(iPrint);
		
		// switch regimes around if observation 1 does not have the highest tau
		// note that the time paths have not been changed; for this the filter needs to be rerun!
//		if (s_mPosteriorMixtureProbabilities[0][0] < s_mPosteriorMixtureProbabilities[0][1])
//		{
//			decl aid1,aid2;
//			aid1 = s_mPosteriorMixtureProbabilities[][0]; s_mPosteriorMixtureProbabilities[][0] = s_mPosteriorMixtureProbabilities[][1]; s_mPosteriorMixtureProbabilities[][1] = aid1;
//			[aid1, aid2] = fGetParameters(vP, iP, s_iEstimateInitialMeans, s_iEstimateInitialCovariances);
//			s_mParameters[0][] = aid2[1][]; s_mParameters[1][] = aid2[0][];
//			vSave = aid1[0] | vecr(s_mParameters[][:iD-1]) | vecr(s_mParameters[][iD:]) | ((sizerc(s_mParameters) + 1 < rows(vP)) ? vP[sizerc(s_mParameters) + 1:] : <>);
//		}

		decl vSE   = zeros(vP);
		decl v_tval= zeros(vP);
		if (dF > 29100) [vSE, v_tval] =	ParStderr(vP);

		// store results for this starting value
		mResults  |=	dF ~ dAICc ~ dBIC ~ dBNg1 ~ dBNg2 ~ dBNg3 ~ dNCH ~ dAICk ~ dNSC ~ dDBI ~ dSSEw
		           ~ ir ~ s_iRanSeed ~vP' ~ RestrictPar(vP)' ~ vSE' ~ v_tval' ~ meanc(s_mPosteriorMixtureProbabilities) ~ 0 ~ sortr(meanc(s_mPosteriorMixtureProbabilities));
		mResults2 |= vecr(s_mPosteriorMixtureProbabilities)';
		

	// save result files
	savemat(s_sName+"_Results.xls", mResults);
	savemat(s_sName+"_vTauMax.xls", vTauMax);

//	if ((dF == maxc(mResults[][0]) && dF > 29190 && ir<1) || s_iRanSeed == 1079) { 
//		savemat(s_sName+"_BestTaus.xls",                       s_vIdt ~ s_mPosteriorMixtureProbabilities);
//		savemat(s_sName+"_BestTaus"+sprint(s_iRanSeed)+".xls", s_vIdt ~ s_mPosteriorMixtureProbabilities);
//		println("!!!!!Found a better maximum!!!!!");
//
//		// plot the time varying parameter path at the new max
//		DrawTVpar(0, s_ammParameters);
//		DrawMCov(0, s_ammParameters);
//		DrawDiagnostic(0, s_ammPITTaus, "PIT");
//		DrawDiagnostic(0, s_ammFilteredTaus, "Filtered");
//		}
	
   	}
	
	// choose result for best starting value and evaluate objective function once more to
	// set time varying parameter path
	decl vResults  = selectifr(mResults,  mResults[][0] .== maxc(mResults[][0]) .? 1 .: 0)[0][];
	decl vResults2 = selectifr(mResults2, mResults[][0] .== maxc(mResults[][0]) .? 1 .: 0)[0][];
 
	s_mPosteriorMixtureProbabilities = reshape(vResults2, rows(s_mPosteriorMixtureProbabilities), columns(s_mPosteriorMixtureProbabilities));
	//vP = vResults[6:5+rows(vP)]';
	//[vEwma, s_mParameters] = fGetParameters(vP, iP, s_iEstimateInitialMeans, s_iEstimateInitialCovariances);
	//fPerformEStep(s_mParameters, vEwma);

	decl mZ = (s_mPosteriorMixtureProbabilities .== fMaxr(s_mPosteriorMixtureProbabilities)*ones(1,iP) .? 1 .: 0);
	savemat(s_sName+"_BestClassification.xls", s_vIdt ~ mZ);
	
//	// plot the time varying parameter path at the end
//	if (iPrint > -100)
//	{	DrawParametric(0, s_ammParameters);
//		DrawTVpar(0, s_ammParameters);
//		DrawMCov(0, s_ammParameters);
//		DrawDiagnostic(0, s_ammPITTaus, "PIT");
//		DrawDiagnostic(0, s_ammFilteredTaus, "Filtered");
//		//ShowDrawWindow();
//	}
	
	return vResults; //~ fVecr3dStructure(s_ammParameters) ~ fVecr3dStructure(s_ammTrueParameters);
	
}




////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        SIMULATION PLOTTING ROUTINES                              //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

fSimulateData1(const iNrFirmsPerMixtureComponent, const iT, const iNrMixtures, const dInverseDegreesOfFreedom, 
		const dCircleRadius, const dCircleMidpointDistance, const iNrLaps, const iPlot) {
// simulate data;
//
// INPUTs:
// iPlot: integer; if 1, then plot the data while generating
//
// NO OUTPUTS

	decl iN = iNrFirmsPerMixtureComponent, iP = iNrMixtures, iD = 2;		  //iN = firms per cluster!
	decl vTime = iNrLaps*M_2PI * range(1,iT) / iT, vMean;
	decl mX, mY, mZ, mAid, mCorr, mChol, iOffset=0;

	// initialize
	mX = mY = mZ = zeros(iP*iN,iT);
	s_ammData = new array[iT];

	for (decl i1=0;i1<iT;++i1) {

		iOffset = 0;	                             //start with first component!
		s_ammData[i1] = zeros(iP*iN, iD);

		for (decl i2=0;i2<iP;++i2) {
	
			vMean =
				(dCircleRadius*cos(((-1)^i2) * vTime[i1]) + i2*dCircleMidpointDistance) |
				(dCircleRadius*sin(((-1)^i2) * vTime[i1]));
			mCorr = <1,0;0,1>;
			mChol = choleski(mCorr)';
			s_ammTrueParameters[i1][i2][] = (vMean | vech(mChol'))';
			mAid  = rann(iN,iD) * mChol;	  //iN x 2
			if (ismissing(dInverseDegreesOfFreedom) || (dInverseDegreesOfFreedom == 0)) {} else {
				// generate Student t
				mAid ./= sqrt( fabs(dInverseDegreesOfFreedom) * ranchi(iN,1,1/fabs(dInverseDegreesOfFreedom)) );
			}

			s_ammData[i1][iOffset:iOffset+iN-1][0] = mX[iOffset:iOffset+iN-1][i1] = vMean[0] + mAid[][0];
			s_ammData[i1][iOffset:iOffset+iN-1][1] = mY[iOffset:iOffset+iN-1][i1] =	vMean[1] + mAid[][1];
			mZ[iOffset:iOffset+iN-1][i1] = i2;
			if (iPlot==1) DrawXMatrix(0,mY[iOffset:iOffset+iN-1][i1]',0,mX[iOffset:iOffset+iN-1][i1]',0,1,2+i2);
			if (iPlot==1) {
				DrawAdjust(ADJ_SYMBOL,2,150);
				decl dmin = 5+dCircleRadius;
				DrawAxis(0,1,0,-dmin,dmin+(iP-1)*dCircleMidpointDistance,-dmin,1,1,0);
				DrawAxis(0,0,0,-dmin,dmin+dCircleRadius,-dmin,1,1,0);
			}
			iOffset += iN;
		}

		if (iPlot==1) {
//			DrawAxis(0,1,0,-5-dCircleRadius,5+dCircleRadius+(iP-1)*dCircleMidpointDistance,,1,1,);
//			DrawAxis(0,1,0,-5-dCircleRadius,5+dCircleRadius,1,0, 0,0);
//			DrawAdjust(ADJ_AREA_X,-5-dCircleRadius,5+dCircleRadius+(iP-1)*dCircleMidpointDistance,1);
//			DrawAdjust(ADJ_AREA_Y,-5-dCircleRadius,5+dCircleRadius,1);
			ShowDrawWindow();
		}
	}

	s_mPosteriorMixtureProbabilities = ones(iN*iP,iP)/iP;

	return {mX, mY, mZ};
	
}




////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        DATA LOADING AND CLEANING ROUTINES                        //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


CleanMV(mX)
//Replaces missings by last known value, then filling forwards for 2008Q1-Q3  
{	decl r,c;
	decl iN = rows(mX), iT = columns(mX);
	//filling in last known value
	for (r=0;r<iN;r++) for (c=   1;c<iT;c +=1) if (mX[r][c] .== .NaN) mX[r][c] = mX[r][c-1];
	//filling in forwards nearest known value for 2008Q1-Q3
	for (r=0;r<iN;r++) for (c=iT-2;c>-1;c -=1) if (mX[r][c] .== .NaN) mX[r][c] = mX[r][c+1];
	return mX;

}

LoadData(const iPlot)
{
	decl iN, iT, mAid, iD, iP, i=0, t=0;

	// load first data matrix
	decl mTA   = loadmat("Data2016/TA.xls")[][1:];	     //total assets	                           -> size
	decl mCET  = loadmat("Data2016/CET.xls")[][1:];	     //total CET1 capital                      -> size
	decl mLev;

	decl mLA   = loadmat("Data2016/L2A.xls")[][1:];	     //loans over assets ratio		           ->  complexity
	decl mRM   = loadmat("Data2016/RM.xls")[][1:];		 //risk mix = (MR + OpR) / CR 			   ->	complexity
	decl mAHT  = loadmat("Data2016/AHT.xls")[][1:];		 //assets held for trading			       ->	complexity
	decl mDHT  = loadmat("Data2016/DHT.xls")[][1:];		 //derivatives held for trading			   ->	complexity
	
	decl mNII  = loadmat("Data2016/NII.xls")[][1:];		 //net interest income / operating revenue   -> activity
	decl mNFC  = loadmat("Data2016/FCI.xls")[][1:];		 //net fees & commissions / operating income -> activity
	decl mTII  = loadmat("Data2016/TII.xls")[][1:];		 //trading income / operating income         -> activity
	decl mRLT  = loadmat("Data2016/RL2L_A.xls")[][1:];	 //retail loans / (retail + corporate loans) -> activity

	decl mDL2L = loadmat("Data2016/DL2L_A2.xls")[][1:];	 //domestic loans over loans ratio    -> geography
	
	decl mL2D  = loadmat("Data2016/L2D.xls")[][1:];	   	 //loans over deposits ratio	           -> funding
	
	decl mOWN  = loadmat("Data2016/OWN.xls")[][1];		 //Ownership structure & status 		     ->	ownership
	decl vIdt  = loadmat("Data2016/TA.xls")[][0];

	//s_mGASX    = loadmat("Data2016/GASX_Q.xls")[][1:2];
	//s_mGASX    = diff0(s_mGASX, 4)';
	s_mGASX    = s_mGASX';
	println("Dimensions of s_mGASX are ",rows(s_mGASX)," x ", columns(s_mGASX) ); 
	s_mGASX   /= 100;  //in absolutes for numerical reasons!

	//s_mB1select = loadmat("Data2016/s_mB1select.xls")[1:][];
	//println("s_mB1select: ", s_mB1select);
	//exit(0);
	
	//reorder chronologically 2008Q1 - 2015Q4, then replace missings with closest available information
		 mTA   = CleanMV(   mTA[][range(31,0,-1)] );	     
		 mCET  = CleanMV(  mCET[][range(31,0,-1)] );
		 mLA   = CleanMV(   mLA[][range(31,0,-1)] );
 		 mRM   = CleanMV(   mRM[][range(31,0,-1)] );
		 mAHT  = CleanMV(  mAHT[][range(31,0,-1)] );
		 mDHT  = CleanMV(  mDHT[][range(31,0,-1)] );
		 mNII  = CleanMV(  mNII[][range(31,0,-1)] );
		 mNFC  = CleanMV(  mNFC[][range(31,0,-1)] );
		 mTII  = CleanMV(  mTII[][range(31,0,-1)] );		 
	     mRLT  = CleanMV(  mRLT[][range(31,0,-1)] );
		 mDL2L = CleanMV( mDL2L[][range(31,0,-1)] );
		 mL2D  = CleanMV(  mL2D[][range(31,0,-1)] );
	
	//transform data to "approx normal"
	mTA   = log(mTA  / 10^6 );        //  
		mTA = (mTA .== -.Inf .? .NaN .: mTA);
	mCET  = (mCET / 10^6 );		   
		mCET = log(mCET .< 0 .? 1 .: mCET);
	mLev = (mTA) - (mCET);

//	mLA   = quann(mLA  /100);
//	 	mLA = (mLA .==   .Inf .?   3 .: mLA);  mLA = (mLA .==  -.Inf .?  -3 .: mLA);
	mLA   =  (mLA  /100);
	mRM   =  log(mRM);
		mRM   = (mRM .<  -4 .?  -4 .: mRM);														    	 
	mAHT  =  (mAHT / 10^6) ./ exp(mTA); 		 
		mAHT = (mAHT .> 1 .? 1 .: mAHT);		 
	mDHT  =  (mDHT / 10^6) ./ exp(mTA); 			//- (mTA);
		mDHT = (mDHT .> 1 .? 1 .: mDHT);		//decl mDHT2 = log (1 + mDHT .* exp(mTA));	


	mNII  = mNII  / 100;
		mNII = (mNII .> 3 .? 3 .: mNII);  mNII = (mNII .< -1.5 .? -1.5 .: mNII);						  
	mNFC  = mNFC  / 100;
		mNFC = (mNFC .> 1.5 .? 1.5 .: mNFC);  mNFC = (mNFC .< -1.0 .? -1.0 .: mNFC);					 
	mTII  = mTII  / 100;

	mRLT  = (mRLT / 100);
		
	mDL2L = quann(mDL2L / 100);
 		mDL2L = (mDL2L .==  .Inf .?  3.5 .: mDL2L);
		mDL2L = (mDL2L .==  -.Inf .?  -3.5 .: mDL2L);  

	mL2D  = mL2D / 100;			   
		
	mOWN  =	 (mOWN - meanc( deleter(mOWN) )) / sqrt( varc(deleter(mOWN)) );
	ranseed(0);
	mOWN += 1 * rann(rows(mOWN),1);
	mOWN  = mOWN * ones(1,32);
		
//	savemat("Data2016/in7/mTA.in7",   mTA  );
//	savemat("Data2016/in7/mLev.in7",  mLev );
//
//	savemat("Data2016/in7/mLA.in7",   mLA  );
//	savemat("Data2016/in7/mRM.in7",   mRM );
//	savemat("Data2016/in7/mAHT.in7",  mAHT);	
//	savemat("Data2016/in7/mDHT.in7",  mDHT);
//	
//	savemat("Data2016/in7/mNII.in7",  mNII);
//	savemat("Data2016/in7/mNFC.in7",  mNII);
//	savemat("Data2016/in7/mTII.in7",  mTII);
//	savemat("Data2016/in7/mRLT.in7",  mRLT);
//	
//	savemat("Data2016/in7/mDL2L.in7", mDL2L );
//
//	savemat("Data2016/in7/mL2D.in7",  mL2D );
//
//	savemat("Data2016/in7/mOWN.in7", mOWN);

	
	// collect into array
	iT  = columns(mTA);
	s_ammData     = new array[iT];
	s_ammPITTaus  = new array[iT];
	s_ammFilteredTaus = new array[iT];
	
	

	for (t=0; t<iT; t++) {		   //assemble data
		mAid = vIdt ~ mTA[][t]  ~ mLev[][t]	 		       //size
		            ~ mLA[][t]  ~ mRM[][t]  ~ mAHT[][t] ~ mDHT[][t]    //complexity	& non-tradional
					~ mNII[][t] ~ mNFC[][t] ~ mTII[][t] ~ mRLT[][t]   //activities
					~ mDL2L[][t] //geography
					~ mL2D[][t]  //funding
					~ mOWN[][t]; //ownership	

		mAid = deleter(mAid);
		s_ammData[t] = mAid[][1:];	 	 
		
		if (t==0) s_vIdt = mAid[][0];	 
		if (t==0) println("* There are ",rows(mAid)," firms"); // at quarter ", t);

		if (iPlot>0)
		{	SetDrawWindow("Data Movie");
			decl mPlot = mAid[][1:];
			for (i=1; i<13; i++) {
				DrawXMatrix(i-1,mPlot[][0]',0,mPlot[][i]',0,1,2);
				DrawAdjust(ADJ_SYMBOL,2,150);
			};
			ShowDrawWindow();
			//savemat("Data2016/NII_Q.in7", mNII);
		}

	};
	
	return 1;
	
}

 
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		                      MAIN                                        //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

main() {

	// initializations
	decl aProgramArguments = arglist();
	decl iNrSimulations = 1, iSim;
	decl vResult, mResult = <>;
	decl dInverseDegreesOfFreedomForSimulation = 0.2, 		dInverseDegreesOfFreedomForEstimation = 1/5; 
	decl iNrMixtureComponentsForSimulation=2,				iNrMixtureComponentsForEstimation=6;
	decl dCircleRadius=1, dCircleMidpointDistance=8, iNrLaps=1;
	decl iN=200, iT=32, iD=2; // iN is number of firms PER MIXTURE COMPONENT; total nr firms is iN*iNrMixtureComponentsForSimulation
	decl iNrStartingValuesForEM = 1, iRanSeed = s_iRanSeed;
	s_iPrint = 10;

	// initialize necessary globals
	s_iEstimateInitialMeans = 0;
	s_iEstimateInitialCovariances = 0;
	s_ammParameters = fInitializeTimeVaryingParameterStructure(iT,iNrMixtureComponentsForEstimation,0.5*iD*(iD+3));
	s_ammTrueParameters = fInitializeTimeVaryingParameterStructure(iT,iNrMixtureComponentsForSimulation,0.5*iD*(iD+3));
	// note that s_mU has the number of firms equal to iN*iNrMixtureComponentsForSimulation,
	// note that s_mU has the number of firms equal to iN*iNrMixtureComponentsForSimulation,
	// even though s_mU is used in the estimation. That is because iN*iNrMixtureComponentsForSimulation
	// is the actual number of firms in the data in the current simulation engine.
	s_mU = fInitializeTimeVaryingParameterStructure(iT,iN*iNrMixtureComponentsForSimulation,iNrMixtureComponentsForSimulation,1);

	// overwrite global choices with program arguments list
	fUseCommandLineArguments(aProgramArguments
		,&iRanSeed
//		,&iNrSimulations
		,&iN
//		,&iT
//		,&iD
		,&dInverseDegreesOfFreedomForSimulation
		,&dInverseDegreesOfFreedomForEstimation
		,&dCircleRadius
		,&dCircleMidpointDistance
//		,&iNrLaps
//		,&iNrStartingValuesForEM
	);
	println("\n\n=================================\nSeed:",iRanSeed,"\n(iN,iT,iD) = (",iN,",",iT,",",iD,")\n(iPsim,iPest) = (",iNrMixtureComponentsForSimulation,",",iNrMixtureComponentsForEstimation,")\n(DFiSim, DFiEst) = (",dInverseDegreesOfFreedomForSimulation,",",dInverseDegreesOfFreedomForEstimation,")\n(Radius,Dist,Laps) = (",dCircleRadius,",",dCircleMidpointDistance,",",iNrLaps,")\n(#Sims,#Starts) = (",iNrSimulations,",",iNrStartingValuesForEM,")\n\n============================\n");

	// run simulations
	if (rows(s_iRanSeed) > 0) {println("Setting the random seed to ",s_iRanSeed); ranseed(s_iRanSeed);}
	for (iSim=0;iSim<iNrSimulations;++iSim) {

		// sign of life
		if (fmod(iSim,1)==0) {
			println("\n==================================");
			println("Estimation/Simulation: ",iSim);
			println("==================================");
		}
		
//		// simulation
//		fSimulateData1(iN, iT, iNrMixtureComponentsForSimulation, dInverseDegreesOfFreedomForSimulation, dCircleRadius, dCircleMidpointDistance, iNrLaps,0);
//		fSimulateData1(iN, iT, iNrMixtureComponentsForSimulation, dInverseDegreesOfFreedomForSimulation, dCircleRadius, dCircleMidpointDistance, iNrLaps,0);
//		fSimulateData1(iN, iT, iNrMixtureComponentsForSimulation, dInverseDegreesOfFreedomForSimulation, dCircleRadius, dCircleMidpointDistance, iNrLaps,0);
//		fSimulateData1(iN, iT, iNrMixtureComponentsForSimulation, dInverseDegreesOfFreedomForSimulation, dCircleRadius, dCircleMidpointDistance, iNrLaps,0);
//		fSimulateData1(iN, iT, iNrMixtureComponentsForSimulation, dInverseDegreesOfFreedomForSimulation, dCircleRadius, dCircleMidpointDistance, iNrLaps,0);

		//load data
		LoadData(0);
		s_sBankNames = BankNames();

		s_sName = "Table7/Dyn_Est_"~sprint(dInverseDegreesOfFreedomForEstimation)~"_"~sprint(iNrMixtureComponentsForEstimation);
		// estimation
		vResult = fRunEmAlgorithm(iNrMixtureComponentsForEstimation, iNrStartingValuesForEM, s_iPrint, dInverseDegreesOfFreedomForEstimation);

		// store results
		mResult |= vResult;
	
	}	

	// save result files
	savemat(s_sName+"_Final.xls", mResult);
	println("mResult: ", mResult); 

	
	// Just print this number to be reminded it is there
	println("\n\nJust to remember it is there: regularization epsilon = ",s_dRegularizeCovariance);	
}

